# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from mock import Mock, patch

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.Block import Block
from qrl.core.State import State
from qrl.core.txs.CoinBase import CoinBase
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.TransactionPool import TransactionPool
from tests.misc.helper import replacement_getTime, set_qrl_dir, get_alice_xmss, get_bob_xmss
from tests.misc.MockHelper.mock_function import MockFunction

logger.initialize_default()


def make_tx(txhash=b'hashbrownies', fee=1, autospec=TransferTransaction, PK=b'publickey', **kwargs):
    return Mock(autospec=autospec, txhash=txhash, fee=fee, PK=PK, **kwargs)


def replacement_from_pbdata(protobuf_tx):
    return protobuf_tx


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestTransactionPool(TestCase):
    """
    TransactionPool sits between incoming Transactions from the network and Blocks.
    First, incoming Transactions are pending Transactions and go into TransactionPool.pending_tx_pool.
    The TxnProcessor has to validate them. Once they are validated, the TxnProcessor puts them into
    TransactionPool.transaction_pool, where they wait to be put into the next mined Block.
    """

    def setUp(self):
        self.txpool = TransactionPool(None)

    def test_add_tx_to_pool(self):
        tx = make_tx()
        result = self.txpool.add_tx_to_pool(tx, 1, replacement_getTime())
        self.assertTrue(result)
        self.assertEqual(len(self.txpool.transactions), 1)

    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', autospec=True)
    def test_add_tx_to_pool_while_full(self, m_is_full_func):
        m_is_full_func.return_value = True
        tx = make_tx()
        result = self.txpool.add_tx_to_pool(tx, 1, replacement_getTime())
        self.assertFalse(result)  # refused to add to the pool
        self.assertEqual(len(self.txpool.transactions), 0)  # remains untouched

    @patch('qrl.core.TransactionPool.config', autospec=True)
    def test_is_full_transaction_pool(self, m_config):
        m_config.user.transaction_pool_size = 2

        result = self.txpool.is_full_transaction_pool()
        self.assertFalse(result)

        tx1 = make_tx(fee=1)
        tx2 = make_tx(fee=2)

        self.txpool.add_tx_to_pool(tx1, 1, replacement_getTime())
        self.txpool.add_tx_to_pool(tx2, 1, replacement_getTime())

        result = self.txpool.is_full_transaction_pool()
        self.assertTrue(result)

    def test_get_tx_index_from_pool(self):
        tx1 = make_tx(txhash=b'red')
        tx2 = make_tx(txhash=b'blue')
        tx3 = make_tx(txhash=b'qrlpink')

        self.txpool.add_tx_to_pool(tx1, 1, replacement_getTime())
        self.txpool.add_tx_to_pool(tx2, 1, replacement_getTime())
        self.txpool.add_tx_to_pool(tx3, 1, replacement_getTime())

        idx = self.txpool.get_tx_index_from_pool(b'qrlpink')
        self.assertEqual(idx, 2)

        idx = self.txpool.get_tx_index_from_pool(b'red')
        self.assertEqual(idx, 0)

        idx = self.txpool.get_tx_index_from_pool(b'ultraviolet')
        self.assertEqual(idx, -1)

    def test_remove_tx_from_pool(self):
        tx1 = make_tx(txhash=b'red')
        tx2 = make_tx(txhash=b'blue')
        tx3 = make_tx(txhash=b'qrlpink')

        self.txpool.add_tx_to_pool(tx1, 1, replacement_getTime())

        # If we try to remove a tx that wasn't there, the transaction pool should be untouched
        self.assertEqual(len(self.txpool.transaction_pool), 1)
        self.txpool.remove_tx_from_pool(tx2)
        self.assertEqual(len(self.txpool.transaction_pool), 1)

        # Now let's remove a tx from the heap. The size should decrease.
        self.txpool.add_tx_to_pool(tx2, 1, replacement_getTime())
        self.txpool.add_tx_to_pool(tx3, 1, replacement_getTime())

        self.assertEqual(len(self.txpool.transaction_pool), 3)
        self.txpool.remove_tx_from_pool(tx2)
        self.assertEqual(len(self.txpool.transaction_pool), 2)

    @patch('qrl.core.TransactionPool.TransactionPool.is_full_pending_transaction_pool', autospec=True)
    def test_update_pending_tx_pool(self, m_is_full_pending_transaction_pool):
        tx1 = make_tx()
        ip = '127.0.0.1'
        m_is_full_pending_transaction_pool.return_value = False

        # Due to the straightforward way the function is written, no special setup is needed to get the tx to go in.
        result = self.txpool.update_pending_tx_pool(tx1, ip)
        self.assertTrue(result)

        # If we try to re-add the same tx to the pending_tx_pool, though, it should fail.
        result = self.txpool.update_pending_tx_pool(tx1, ip)
        self.assertFalse(result)

    @patch('qrl.core.TransactionPool.TransactionPool.is_full_pending_transaction_pool', autospec=True)
    def test_update_pending_tx_pool_tx_already_validated(self, m_is_full_pending_transaction_pool):
        """
        If the tx is already in TransactionPool.transaction_pool, do not add it to pending_tx_pool.
        """
        tx1 = make_tx()
        ip = '127.0.0.1'
        m_is_full_pending_transaction_pool.return_value = False

        self.txpool.add_tx_to_pool(tx1, 1, replacement_getTime())

        result = self.txpool.update_pending_tx_pool(tx1, ip)
        self.assertFalse(result)

    @patch('qrl.core.TransactionPool.TransactionPool.is_full_pending_transaction_pool', autospec=True)
    def test_update_pending_tx_pool_is_full_already(self, m_is_full_pending_transaction_pool):
        tx1 = make_tx()
        ip = '127.0.0.1'
        m_is_full_pending_transaction_pool.return_value = True

        result = self.txpool.update_pending_tx_pool(tx1, ip)
        self.assertFalse(result)

    @patch('qrl.core.TransactionPool.logger')
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_pending_transaction_pool', autospec=True)
    def test_update_pending_tx_pool_rejects_coinbase_txs(self, m_is_full_pending_transaction_pool, m_logger):
        tx1 = CoinBase()
        ip = '127.0.0.1'
        m_is_full_pending_transaction_pool.return_value = False

        result = self.txpool.update_pending_tx_pool(tx1, ip)
        self.assertFalse(result)

    @patch('qrl.core.TransactionPool.config', autospec=True)
    def test_is_full_pending_transaction_pool(self, m_config):
        """
        pending_transaction_pool_size is 3, and pending_transaction_pool_reserve is subtracted out of that, so it's 2.
        Trying to add in 3 transactions with ignore_reserve=True will fail, but if ignore_reserve=False, it will go in.
        However, after that, adding even more transactions will always fail.
        """
        m_config.user.pending_transaction_pool_size = 3
        m_config.user.pending_transaction_pool_reserve = 1

        tx4 = make_tx(txhash=b'red')
        tx1 = make_tx(txhash=b'green')
        tx3 = make_tx(txhash=b'blue')
        tx2 = make_tx(txhash=b'pink')
        ip = '127.0.0.1'

        self.txpool.update_pending_tx_pool(tx1, ip)
        self.txpool.update_pending_tx_pool(tx2, ip)
        result = self.txpool.update_pending_tx_pool(tx3, ip, ignore_reserve=True)
        self.assertFalse(result)
        result = self.txpool.update_pending_tx_pool(tx3, ip, ignore_reserve=False)
        self.assertTrue(result)

        result = self.txpool.update_pending_tx_pool(tx4, ip, ignore_reserve=True)
        self.assertFalse(result)
        result = self.txpool.update_pending_tx_pool(tx4, ip, ignore_reserve=False)
        self.assertFalse(result)

    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def test_get_pending_transaction(self):
        """
        Getting a pending transaction also removes it from the TransactionPool.
        Because it may return a single None, or two variables, a funny hack is used in TxnProcessor where the return
        from this function is stored in one variable then unpacked later if it is not None.
        """
        tx1 = make_tx()
        ip = '127.0.0.1'
        self.txpool.update_pending_tx_pool(tx1, ip)

        self.assertEqual(len(self.txpool.pending_tx_pool_hash), 1)
        tx_timestamp = self.txpool.get_pending_transaction()
        self.assertEqual(tx_timestamp[0], tx1)
        self.assertEqual(len(self.txpool.pending_tx_pool_hash), 0)

        tx_timestamp = self.txpool.get_pending_transaction()
        self.assertIsNone(tx_timestamp)

    @patch('qrl.core.TransactionPool.logger')
    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', return_value=make_tx())
    @patch('qrl.core.TransactionPool.TransactionPool.add_tx_to_pool', return_value=True)
    def test_add_tx_from_block_to_pool(self, m_add_tx_to_pool, m_from_pbdata, m_logger):
        m_block = Mock(autospec=Block, block_number=5, headerhash=b'test block header')
        m_block.transactions = [CoinBase(), make_tx(), make_tx()]

        self.txpool.add_tx_from_block_to_pool(m_block, 5)

        self.assertEqual(m_add_tx_to_pool.call_count, 2)  # 2 because the function ignores the Coinbase tx

        # If there is a problem adding to the tx_pool, the logger should be invoked.
        m_add_tx_to_pool.return_value = False
        self.txpool.add_tx_from_block_to_pool(m_block, 5)
        m_logger.warning.assert_called()

    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', new=replacement_from_pbdata)
    def test_remove_tx_in_block_from_pool(self):
        m_block = Mock(autospec=Block)
        tx1 = make_tx(name='Mock TX 1', ots_key=1, PK=b'pk')
        tx2 = make_tx(name='Mock TX 2', ots_key=2, PK=b'pk')
        m_block.transactions = [CoinBase(), tx1, tx2]

        # To remove the tx from the pool we have to add it first!
        self.txpool.add_tx_to_pool(tx1, 5)
        self.txpool.add_tx_to_pool(tx2, 5)
        self.assertEqual(len(self.txpool.transaction_pool), 2)

        self.txpool.remove_tx_in_block_from_pool(m_block)
        self.assertEqual(len(self.txpool.transaction_pool), 0)

    @patch('qrl.core.TransactionInfo.config', autospec=True)
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', return_value=False)
    def test_check_stale_txn(self, m_is_full_transaction_pool, m_config):
        """
        Stale Transactions are Transactions that were supposed to go into block 5, but for some reason didn't make it.
        They languish in TransactionPool until check_stale_txn() checks the Pool and updates the tx_info to make them
        go into a higher block.
        For each stale transaction, P2PFactory.broadcast_tx() will be called.
        """

        # Redefine at what point should txs be considered stale
        m_config.user.stale_transaction_threshold = 2
        bob_xmss = get_bob_xmss(4)
        alice_xmss = get_alice_xmss(4)

        tx1 = TransferTransaction.create(addrs_to=[bob_xmss.address], amounts=[1000000], fee=1, xmss_pk=alice_xmss.pk)
        tx1.sign(alice_xmss)
        tx2 = TransferTransaction.create(addrs_to=[bob_xmss.address], amounts=[10000], fee=1, xmss_pk=alice_xmss.pk)
        tx2.sign(alice_xmss)
        m_broadcast_tx = Mock(name='Mock Broadcast TX function (in P2PFactory)')
        self.txpool.add_tx_to_pool(tx1, 5)
        self.txpool.add_tx_to_pool(tx2, 5)
        self.txpool.set_broadcast_tx(m_broadcast_tx)

        with set_qrl_dir('no_data'):
            state = State()
            self.txpool.check_stale_txn(state, 8)

            self.assertEqual(m_broadcast_tx.call_count, 0)

            m = MockFunction()
            bob_address_state = AddressState.get_default(bob_xmss.address)
            bob_address_state.pbdata.balance = 1000000000000
            m.put(bob_xmss.address, bob_address_state)
            state.get_address_state = m.get
            tx3 = TransferTransaction.create(addrs_to=[alice_xmss.address], amounts=[10000], fee=1,
                                             xmss_pk=bob_xmss.pk)
            tx3.sign(bob_xmss)
            self.txpool.add_tx_to_pool(tx3, 5)
            self.txpool.check_stale_txn(state, 8)

            self.assertEqual(m_broadcast_tx.call_count, 1)


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestTransactionPoolRemoveTxInBlockFromPool(TestCase):
    """
    Up until 4096 (max_ots_tracking_index), the state of each OTS index USED/UNUSED is stored in a bitfield.
    Default height of wallet is 12, so 2^12 = 4096 obviously
    Above that however, the network only keeps track of the last used OTS index as a number. So the next
    tx.ots_index must be 4096 < ots_index < network_ots_index_counter (AddressState.ots_counter).

    Suppose you have a Block with two Transactions from the same public address in it, with ots_index=4098 and 4099.
    If TransactionPool has 4097, it should be invalidated because 4098 is already used and we are on an counter
    method of keeping track of OTS indexes.
    Of course, 4098 and 4099 also have to be deleted.
    """

    @patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
    def setUp(self):

        self.txpool = TransactionPool(None)

        self.tx_3907 = make_tx(name='Mock TX 3907', txhash=b'h3907', ots_key=3907)

        self.tx_4095 = make_tx(name='Mock TX 4095', txhash=b'h4095', ots_key=4095)
        self.tx_4096 = make_tx(name='Mock TX 4096', txhash=b'h4096', ots_key=4096)
        self.tx_4097 = make_tx(name='Mock TX 4097', txhash=b'h4097', ots_key=4097)
        self.tx_4098 = make_tx(name='Mock TX 4098', txhash=b'h4098', ots_key=4098)
        self.tx_4099 = make_tx(name='Mock TX 4099', txhash=b'h4099', ots_key=4099)
        self.tx_4100 = make_tx(name='Mock TX 4100', txhash=b'h4100', ots_key=4100)
        self.tx_4200 = make_tx(name='Mock TX 4200', txhash=b'h4200', ots_key=4200)

        # To remove the tx from the pool we have to add it first!
        self.txpool.add_tx_to_pool(self.tx_4095, 5)
        self.txpool.add_tx_to_pool(self.tx_4096, 5)
        self.txpool.add_tx_to_pool(self.tx_4097, 5)
        self.txpool.add_tx_to_pool(self.tx_4098, 5)
        self.txpool.add_tx_to_pool(self.tx_4099, 5)
        self.txpool.add_tx_to_pool(self.tx_4100, 5)
        self.txpool.add_tx_to_pool(self.tx_4200, 5)

    @patch('qrl.core.TransactionPool.config', autospec=True)
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', return_value=False)
    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', new=replacement_from_pbdata)
    def test_block_4098_4099(self, m_is_full_transaction_pool, m_config):
        """
        TxPool = [4095-4100, 4200]
        Block = [4098, 4099]
        TxPool Afterwards = [4095, 4100, 4200]
        """
        # Ensure that a "large OTS index" is 4096
        m_config.dev.max_ots_tracking_index = 4096

        m_block = Mock(autospec=Block)
        m_block.transactions = [CoinBase(), self.tx_4098, self.tx_4099]

        self.txpool.remove_tx_in_block_from_pool(m_block)
        txs_in_txpool = [t[1].transaction for t in self.txpool.transaction_pool]

        self.assertEqual(len(self.txpool.transaction_pool), 3)
        self.assertNotIn(self.tx_4097, txs_in_txpool)
        self.assertNotIn(self.tx_4098, txs_in_txpool)
        self.assertNotIn(self.tx_4099, txs_in_txpool)

        self.assertIn(self.tx_4095, txs_in_txpool)
        self.assertIn(self.tx_4100, txs_in_txpool)
        self.assertIn(self.tx_4200, txs_in_txpool)

    @patch('qrl.core.TransactionPool.config', autospec=True)
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', return_value=False)
    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', new=replacement_from_pbdata)
    def test_txpool_3907_block_4098_4099(self, m_is_full_transaction_pool, m_config):
        """
        TxPool = [3907, 4095-4100, 4200]
        Block = [4098, 4099]
        TxPool Afterwards = [3907, 4095, 4100, 4200]
        """
        # Ensure that a "large OTS index" is 4096
        m_config.dev.max_ots_tracking_index = 4096

        m_block = Mock(autospec=Block)
        m_block.transactions = [CoinBase(), self.tx_4098, self.tx_4099]

        self.txpool.add_tx_to_pool(self.tx_3907, 5)

        self.txpool.remove_tx_in_block_from_pool(m_block)
        txs_in_txpool = [t[1].transaction for t in self.txpool.transaction_pool]

        # 3907 should also be in the Pool since it is exempt from the counter
        self.assertEqual(len(self.txpool.transaction_pool), 4)
        self.assertNotIn(self.tx_4097, txs_in_txpool)
        self.assertNotIn(self.tx_4098, txs_in_txpool)
        self.assertNotIn(self.tx_4099, txs_in_txpool)

        self.assertIn(self.tx_3907, txs_in_txpool)
        self.assertIn(self.tx_4095, txs_in_txpool)
        self.assertIn(self.tx_4100, txs_in_txpool)
        self.assertIn(self.tx_4200, txs_in_txpool)

    @patch('qrl.core.TransactionPool.config', autospec=True)
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', return_value=False)
    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', new=replacement_from_pbdata)
    def test_block_4200(self, m_is_full_transaction_pool, m_config):
        """
        TxPool = [3907, 4095-4100, 4200]
        Block = [4200]
        TxPool Afterwards = [3907, 4095]
        """
        # Ensure that a "large OTS index" is 4096
        m_config.dev.max_ots_tracking_index = 4096

        m_block = Mock(autospec=Block)
        m_block.transactions = [CoinBase(), self.tx_4200]

        self.txpool.add_tx_to_pool(self.tx_3907, 5)

        self.txpool.remove_tx_in_block_from_pool(m_block)
        txs_in_txpool = [t[1].transaction for t in self.txpool.transaction_pool]

        self.assertEqual(len(self.txpool.transaction_pool), 2)
        self.assertIn(self.tx_3907, txs_in_txpool)
        self.assertIn(self.tx_4095, txs_in_txpool)

    @patch('qrl.core.TransactionPool.config', autospec=True)
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', return_value=False)
    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', new=replacement_from_pbdata)
    def test_txpool_4095_4096_4097_otherppl_block_4098_4099(self, m_is_full_transaction_pool, m_config):
        """
        TxPool = [4096-4100, 4200, 4095-4097_otherppl]
        Block = [4200]
        TxPool Afterwards = [4095, 4095-4097_otherppl]
        """
        # Ensure that a "large OTS index" is 4096
        m_config.dev.max_ots_tracking_index = 4096

        m_block = Mock(autospec=Block)
        m_block.transactions = [CoinBase(), self.tx_4200]

        tx_other_4095 = make_tx(name='Mock TX 4095', txhash=b'h4095_other', ots_key=4095, PK='otherppl')
        tx_other_4096 = make_tx(name='Mock TX 4096', txhash=b'h4096_other', ots_key=4096, PK='otherppl')
        tx_other_4097 = make_tx(name='Mock TX 4097', txhash=b'h4097_other', ots_key=4097, PK='otherppl')

        self.txpool.add_tx_to_pool(tx_other_4095, 5)
        self.txpool.add_tx_to_pool(tx_other_4096, 5)
        self.txpool.add_tx_to_pool(tx_other_4097, 5)

        self.txpool.remove_tx_in_block_from_pool(m_block)
        txs_in_txpool = [t[1].transaction for t in self.txpool.transaction_pool]

        self.assertEqual(len(self.txpool.transaction_pool), 4)
        self.assertIn(self.tx_4095, txs_in_txpool)
        self.assertIn(tx_other_4095, txs_in_txpool)
        self.assertIn(tx_other_4096, txs_in_txpool)
        self.assertIn(tx_other_4097, txs_in_txpool)

    @patch('qrl.core.TransactionPool.config', autospec=True)
    @patch('qrl.core.TransactionPool.TransactionPool.is_full_transaction_pool', return_value=False)
    @patch('qrl.core.txs.Transaction.Transaction.from_pbdata', new=replacement_from_pbdata)
    def test_block_1000(self, m_is_full_transaction_pool, m_config):
        """
        TxPool = [4095-4100, 4200]
        Block = [1000]
        TxPool Afterwards = [4095-4100, 4200]
        """
        # Ensure that a "large OTS index" is 4096
        m_config.dev.max_ots_tracking_index = 4096

        tx_1000 = make_tx(name='Mock TX 1000', txhash=b'h1000', ots_key=1000)
        m_block = Mock(autospec=Block)
        m_block.transactions = [CoinBase(), tx_1000]
        self.assertEqual(7, len(self.txpool.transaction_pool))

        self.txpool.remove_tx_in_block_from_pool(m_block)

        txs_in_txpool = [t[1].transaction for t in self.txpool.transaction_pool]
        self.assertEqual(7, len(txs_in_txpool))



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

import sys
from time import time

from ntplib import NTPClient

from qrl.core import config
from qrl.core.misc import logger

NTP_VERSION = 3
NTP_RETRIES = 6
drift = None
last_refresh = 0


def get_ntp_response():
    for retry in range(NTP_RETRIES):
        ntp_server = config.user.ntp_servers[retry % len(config.user.ntp_servers)]
        try:
            ntp_client = NTPClient()
            response = ntp_client.request(ntp_server, version=NTP_VERSION, timeout=config.user.ntp_request_timeout)
        except Exception as e:
            logger.warning(e)
            continue
        return response

    # FIXME: Provide some proper clean before exiting
    logger.fatal("Could not contact NTP servers after %d retries", NTP_RETRIES)
    sys.exit(-1)


def getNTP():
    ntp_timestamp = 0
    response = get_ntp_response()
    if response:
        ntp_timestamp = int(response.tx_time)
    return ntp_timestamp


def setDrift():
    global drift, last_refresh
    response = get_ntp_response()
    if not response:
        return response
    drift = response.offset
    last_refresh = drift + int(time())


def getTime():
    """
    :return:
    :rtype: int

    >>> getTime() is not None
    True
    """
    global drift, last_refresh

    if drift is None:
        setDrift()

    curr_time = int(drift + time())

    if curr_time - last_refresh > config.user.ntp_refresh:
        setDrift()

    return curr_time



from pyqrllib.pyqrllib import bin2hstr

from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction


class CoinBase(Transaction):
    """
    CoinBase is the type of transaction to credit the block_reward to
    the stake selector who created the block.
    """

    def __init__(self, protobuf_transaction=None):
        super(CoinBase, self).__init__(protobuf_transaction)

    @property
    def addr_to(self):
        return self._data.coinbase.addr_to

    @property
    def amount(self):
        return self._data.coinbase.amount

    def get_data_bytes(self):
        # nonce only added to the hashable bytes of CoinBase
        return self.master_addr + \
               self.addr_to + \
               self.nonce.to_bytes(8, byteorder='big', signed=False) + \
               self.amount.to_bytes(8, byteorder='big', signed=False)

    @staticmethod
    def create(amount, miner_address, block_number):
        transaction = CoinBase()
        transaction._data.master_addr = config.dev.coinbase_address
        transaction._data.coinbase.addr_to = miner_address
        transaction._data.coinbase.amount = amount
        transaction._data.nonce = block_number + 1
        transaction._data.transaction_hash = transaction.get_data_hash()

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def update_mining_address(self, mining_address: bytes):
        self._data.coinbase.addr_to = mining_address
        self._data.transaction_hash = self.get_data_hash()

    def _coinbase_filter(self):
        pass

    def _get_allowed_access_types(self):
        # FIXME: 0 and 1 are not clear..
        return [0, 1]

    def _get_master_address(self):
        return self.addr_to

    def _validate_custom(self):
        if self.fee != 0:
            logger.warning('Fee for coinbase transaction should be 0')
            return False

        return True

    # noinspection PyBroadException
    def validate_extended(self, block_number: int):
        if self.master_addr != config.dev.coinbase_address:
            logger.warning('Master address doesnt match with coinbase_address')
            logger.warning('%s %s', bin2hstr(self.master_addr), bin2hstr(config.dev.coinbase_address))
            return False

        if not AddressState.address_is_valid(self.addr_to):
            logger.warning('Invalid address addr_from: %s addr_to: %s', bin2hstr(self.master_addr), bin2hstr(self.addr_to))
            return False

        if self.nonce != block_number + 1:
            logger.warning('Nonce %s doesnt match with block_number %s', self.nonce, block_number)
            return False

        return self._validate_custom()

    def apply_state_changes(self, addresses_state):
        if self.addr_to in addresses_state:
            addresses_state[self.addr_to].balance += self.amount
            addresses_state[self.addr_to].transaction_hashes.append(self.txhash)

        addr_from = config.dev.coinbase_address

        if self.master_addr in addresses_state:
            addresses_state[self.master_addr].balance -= self.amount
            addresses_state[self.master_addr].transaction_hashes.append(self.txhash)
            addresses_state[addr_from].increase_nonce()

    def revert_state_changes(self, addresses_state, chain_manager):
        if self.addr_to in addresses_state:
            addresses_state[self.addr_to].balance -= self.amount
            addresses_state[self.addr_to].transaction_hashes.remove(self.txhash)

        addr_from = config.dev.coinbase_address

        if self.master_addr in addresses_state:
            addresses_state[self.master_addr].balance += self.amount
            addresses_state[self.master_addr].transaction_hashes.remove(self.txhash)
            addresses_state[addr_from].decrease_nonce()

    def set_affected_address(self, addresses_set: set):
        addresses_set.add(self.master_addr)
        addresses_set.add(self.addr_to)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from qrl.core.misc import logger
from tests.blockchain.MockedBlockchain import MockedBlockchain

logger.initialize_default()


class TestBlockchain(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestBlockchain, self).__init__(*args, **kwargs)

    def test_10blocks(self):
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            last_block = mock_blockchain.qrlnode.get_block_last()
            self.assertEqual(10, last_block.block_number)

    def test_11blocks(self):
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            last_block = mock_blockchain.qrlnode.get_block_last()
            self.assertEqual(10, last_block.block_number)

            mock_blockchain.add_new_block()
            last_block = mock_blockchain.qrlnode.get_block_last()
            self.assertEqual(11, last_block.block_number)

    def test_sequential_index(self):
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            block = mock_blockchain.qrlnode.get_block_last()

            while block is not None:
                prev_block = mock_blockchain.qrlnode.get_block_from_hash(block.prev_headerhash)
                if prev_block:
                    self.assertEqual(block.block_number - 1, prev_block.block_number)
                block = prev_block

    def test_get_by_index(self):
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            block = mock_blockchain.qrlnode.get_block_last()

            while block is not None:
                prev_block = mock_blockchain.qrlnode.get_block_from_hash(block.prev_headerhash)
                if prev_block:
                    self.assertEqual(prev_block.block_number + 1, block.block_number)

                    block_from_index = mock_blockchain.qrlnode.get_block_from_index(block.block_number)
                    self.assertEqual(block._data, block_from_index._data)

                block = prev_block

    def test_MockChain_forking(self):
        number_blocks = 10
        with MockedBlockchain.create(number_blocks) as mock_blockchain:
            block_5 = mock_blockchain.qrlnode.get_block_from_index(9)
            block_6b = mock_blockchain.create_block(block_5.headerhash)
            mock_blockchain.add_block(block_6b)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from mock import Mock, patch

from qrl.core.misc import logger
from qrl.core.processors.TxnProcessor import TxnProcessor
from qrl.core.ChainManager import ChainManager
from qrl.core.State import State
from qrl.core.AddressState import AddressState
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.TransactionPool import TransactionPool
from tests.misc.helper import replacement_getTime
from qrl.core.p2p.p2pfactory import P2PFactory

logger.initialize_default()


def make_tx(txhash=b'hashbrownies', fee=1, autospec=TransferTransaction, PK=b'publickey', **kwargs):
    return Mock(autospec=autospec, txhash=txhash, fee=fee, PK=PK, **kwargs)


@patch('qrl.core.processors.TxnProcessor.logger')
@patch('qrl.core.txs.Transaction.Transaction.get_slave')
class TestTxnProcessor(TestCase):
    def setUp(self):
        m_state = Mock(name='A Mock State', autospec=State)
        m_state.get_address_state.return_value = Mock(name='A Mock AddressState', autospec=AddressState)

        self.chain_manager = Mock(autospec=ChainManager)
        self.chain_manager._state = m_state

        tx_attrs = {
            'validate.return_value': True,  # Custom validation for different Transaction Types
            'validate_extended.return_value': True,  # Master/slave XMSS tree validation; balance & fee, OTS key reuse
            'validate_transaction_pool.return_value': True  # checks for OTS key reuse within TransactionPool only
        }
        self.tx1 = make_tx(name='Mock TX 1', **tx_attrs)
        self.tx2 = make_tx(name='Mock TX 2', **tx_attrs)
        self.tx3 = make_tx(name='Mock TX 3', **tx_attrs)
        self.tx4 = make_tx(name='Mock TX 4', **tx_attrs)

        self.m_txpool = Mock(autospec=TransactionPool)
        self.m_txpool.get_pending_transaction.side_effect = [(self.tx1, replacement_getTime()),
                                                             (self.tx2, replacement_getTime()),
                                                             (self.tx3, replacement_getTime()),
                                                             (self.tx4, replacement_getTime())]

        self.m_broadcast_tx = Mock(autospec=P2PFactory.broadcast_tx)
        self.txnprocessor = TxnProcessor(chain_manager=self.chain_manager,
                                         transaction_pool_obj=self.m_txpool,
                                         broadcast_tx=self.m_broadcast_tx)

    def test_txnprocessor_all_valid(self, m_get_slave, m_logger):
        # Transaction.get_slave() gives you the slave's Qaddress, if the TXN is signed by a slave XMSS tree.
        # If it's a normal TXN signed by the master XMSS tree, it returns None. Since we mocked out validate_extended(),
        # it doesn't really matter what we set here. It's just to make things explicit. Also because class-level patch
        # cannot extend into the setUp() function, only the test_* functions.
        m_get_slave.return_value = b'PUBLICKEY'

        tx_results = [t for t in self.txnprocessor]

        self.assertEqual([True, True, True, True], tx_results)
        self.assertEqual(4, self.m_txpool.add_tx_to_pool.call_count)
        self.assertEqual(4, self.m_broadcast_tx.call_count)

    def test_txnprocessor_tx_validate_fail(self, m_get_slave, m_logger):
        m_get_slave.return_value = None
        self.tx1.validate.return_value = False

        tx_results = [t for t in self.txnprocessor]

        self.assertEqual([False, True, True, True], tx_results)
        self.assertEqual(3, self.m_txpool.add_tx_to_pool.call_count)
        self.assertEqual(3, self.m_broadcast_tx.call_count)

    def test_txnprocessor_tx_validate_extended_fail(self, m_get_slave, m_logger):
        m_get_slave.return_value = None
        self.tx2.validate_extended.return_value = False
        self.tx3.validate_extended.return_value = False

        tx_results = [t for t in self.txnprocessor]

        m_logger.info.assert_called()
        self.assertEqual([True, False, False, True], tx_results)
        self.assertEqual(2, self.m_txpool.add_tx_to_pool.call_count)
        self.assertEqual(2, self.m_broadcast_tx.call_count)

    def test_txnprocessor_tx_validate_transaction_pool_fail(self, m_get_slave, m_logger):
        m_get_slave.return_value = None
        self.tx3.validate_transaction_pool.return_value = False
        self.tx4.validate_transaction_pool.return_value = False

        tx_results = [t for t in self.txnprocessor]

        m_logger.info.assert_called()
        self.assertEqual([True, True, False, False], tx_results)
        self.assertEqual(2, self.m_txpool.add_tx_to_pool.call_count)
        self.assertEqual(2, self.m_broadcast_tx.call_count)

    def test_txnprocessor_tx_all_failure_modes(self, m_get_slave, m_logger):
        m_get_slave.return_value = None
        self.tx2.validate.return_value = False
        self.tx3.validate_extended.return_value = False
        self.tx4.validate_transaction_pool.return_value = False

        tx_results = [t for t in self.txnprocessor]

        m_logger.info.assert_called()
        self.assertEqual([True, False, False, False], tx_results)
        self.assertEqual(1, self.m_txpool.add_tx_to_pool.call_count)
        self.assertEqual(1, self.m_broadcast_tx.call_count)

    def test_empty(self, m_get_slave, m_logger):
        m_get_slave.return_value = None
        self.m_txpool.get_pending_transaction.side_effect = None
        self.m_txpool.get_pending_transaction.return_value = None

        tx_results = [t for t in self.txnprocessor]

        self.assertEqual([], tx_results)
        self.m_txpool.add_tx_to_pool.assert_not_called()
        self.m_broadcast_tx.assert_not_called()



from unittest import TestCase

from mock import Mock, patch, create_autospec, MagicMock
from pyqryptonight.pyqryptonight import StringToUInt256

from qrl.core.AddressState import AddressState
from qrl.core.Block import Block
from qrl.core.misc import logger
from qrl.core.ChainManager import ChainManager
from qrl.core.Miner import Miner
from qrl.core.TransactionPool import TransactionPool
from qrl.core.misc.helper import parse_qaddress
from qrl.core.node import POW
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.txs.CoinBase import CoinBase
from qrl.core.txs.TransferTransaction import TransferTransaction
from tests.misc.helper import get_alice_xmss, get_bob_xmss, replacement_getTime

logger.initialize_default()


alice = get_alice_xmss()
bob = get_bob_xmss()


@patch('qrl.core.Miner.logger')
@patch('qrl.core.misc.ntp.getTime')
class TestMiner(TestCase):
    def setUp(self):
        self.time = 1526830525
        self.m_mining_qaddress = alice.qaddress
        self.m_mining_address = parse_qaddress(self.m_mining_qaddress)

        self.chain_manager = Mock(spec=ChainManager)
        self.parent_block = Block()
        self.parent_difficulty = StringToUInt256('0')  # tuple (0,0,0,0,0...) length 32

        self.m_pre_block_logic = Mock(spec=POW.pre_block_logic, name='hello')
        m_add_unprocessed_txn_fn = create_autospec(P2PFactory.add_unprocessed_txn)
        mining_thread_count = 1

        self.miner = Miner(self.m_pre_block_logic,
                           self.m_mining_address,
                           self.chain_manager,
                           mining_thread_count,
                           m_add_unprocessed_txn_fn)

        self.txpool = Mock(spec=TransactionPool)
        self.txpool.transactions = []

    def test_prepare_next_unmined_block_template_works(self, m_getTime, m_logger):
        """
        All the setup stuff you need before you actually mine a block goes here.
        It's broken out into this function, because if you have a mining pool, this function prepares the
        getblocktemplate for the pool.
        """
        m_getTime.return_value = self.time

        self.chain_manager.get_measurement.return_value = 60
        self.txpool.transactions = []

        self.assertIsNone(self.miner._current_difficulty)
        self.assertIsNone(self.miner._current_target)
        self.assertIsNone(self.miner._measurement)
        self.miner.prepare_next_unmined_block_template(self.m_mining_address,
                                                       self.txpool,
                                                       self.parent_block,
                                                       self.parent_difficulty)

        self.assertEqual(self.miner._current_difficulty, StringToUInt256('2'))
        self.assertEqual(self.miner._current_target, StringToUInt256(
            '115792089237316195423570985008687907853269984665640564039457584007913129639807'))
        self.assertEqual(self.miner._measurement, 60)  # because we set it earlier in this test

    def test_prepare_next_unmined_block_template_exception(self, m_getTime, m_logger):
        """
        If this function should throw an exception, nothing should happen except a call to the logger.
        """
        m_getTime.return_value = self.time

        self.chain_manager.get_measurement.side_effect = ValueError
        self.txpool.transactions = []

        self.assertIsNone(self.miner._current_difficulty)
        self.assertIsNone(self.miner._current_target)
        self.assertIsNone(self.miner._measurement)
        self.miner.prepare_next_unmined_block_template(self.m_mining_address,
                                                       self.txpool,
                                                       self.parent_block,
                                                       self.parent_difficulty)

        self.assertIsNone(self.miner._current_difficulty)
        self.assertIsNone(self.miner._current_target)
        self.assertIsNone(self.miner._measurement)
        m_logger.warning.assert_called_once()
        m_logger.exception.assert_called_once()

    def test_start_mining_works(self, m_getTime, m_logger):
        m_getTime.return_value = self.time

        # Do prepare_next_unmined_block_template()'s job
        self.miner._mining_block = Block()
        # From sample run of test_prepare_next_unmined_block_template_works()
        self.miner._measurement = 60
        self.miner._current_difficulty = StringToUInt256('0')
        self.miner._current_target = \
            StringToUInt256('115792089237316195423570985008687907853269984665640564039457584007913129639807')

        # start() is from Qryptominer, let's not actually mine in a test
        with patch('qrl.core.Miner.Miner.start', spec=True) as m_start:
            self.miner.start_mining(self.parent_block, self.parent_difficulty)
            m_start.assert_called_once()

    def test_get_block_to_mine_no_existing_block_being_mined_upon(self, m_getTime, m_logger):
        """
        This function takes a Qaddress, and returns a blob for the miner/mining pool to work on.
        It makes sure we have a block we're trying to mine and that the coinbase points to the Qaddress.
        In this test we  check that:
        1. If we don't have a block we're trying to mine on, it generates one on the fly.
        """
        m_getTime.return_value = 1526830525
        self.miner._current_difficulty = StringToUInt256('1')

        blob, difficulty = self.miner.get_block_to_mine(self.m_mining_qaddress.encode(), self.txpool, self.parent_block,
                                                        self.parent_difficulty)

        self.assertEqual(difficulty, 1)  # because self.miner._current_difficulty was set above
        self.assertEqual(blob,
                         '0014db80611fbf16e342a2afb8b77b1f513f9db21de3ff905c0c27ea0078c489248f37f9e2a22400000000000000000000000000000000004bfaabbf147f985be702a373183be1be77100b24')  # noqa

    def test_get_block_to_mine_not_mining_upon_last_block(self, m_getTime, m_logger):
        """
        In this test we  check that:
        2. If we aren't mining upon the last block, it regenerates the blocktemplate.
        """
        m_getTime.return_value = 1526830525
        self.miner._current_difficulty = StringToUInt256('1')
        m_mining_block = Mock(autospec=Block)
        m_mining_block.mining_blob = b'big_bad_blob'
        m_mining_block.prev_headerhash = b'nothing should be equal to this'
        self.miner._mining_block = m_mining_block

        blob, difficulty = self.miner.get_block_to_mine(self.m_mining_qaddress.encode(), self.txpool, self.parent_block,
                                                        self.parent_difficulty)

        self.assertEqual(difficulty, 1)  # because self.miner._current_difficulty was set above
        self.assertEqual(blob,
                         '0014db80611fbf16e342a2afb8b77b1f513f9db21de3ff905c0c27ea0078c489248f37f9e2a22400000000000000000000000000000000004bfaabbf147f985be702a373183be1be77100b24')  # noqa

    def test_get_block_to_mine_perfect_block_no_changes(self, m_getTime, m_logger):
        """
        In this test, we check that the function makes no changes to the block if the coinbase addr is our addr,
        and we are on the latest block.
        """
        m_coinbase = Mock(autospec=CoinBase, name='I am a Coinbase')
        m_coinbase.coinbase.addr_to = self.m_mining_address

        m_parent_block = Mock(autospec=Block, name='mock parent_block')
        m_parent_block.transactions = [m_coinbase]

        m_mining_block = Mock(autospec=Block, name='mock _mining_block')
        m_mining_block.transactions = [m_coinbase]
        m_mining_block.mining_blob = b'this is the blob you should iterate the nonce upon'

        self.miner._mining_block = m_mining_block
        self.miner._current_difficulty = StringToUInt256('1')

        m_parent_block.headerhash = b'block_headerhash'
        m_mining_block.prev_headerhash = b'block_headerhash'
        blob, difficulty = self.miner.get_block_to_mine(self.m_mining_qaddress.encode(), self.txpool, m_parent_block,
                                                        self.parent_difficulty)

        self.assertEqual(blob,
                         '746869732069732074686520626c6f6220796f752073686f756c64206974657261746520746865206e6f6e63652075706f6e')
        self.assertEqual(difficulty, 1)

    def test_get_block_to_mine_we_have_a_block_in_mind(self, m_getTime, m_logger):
        """
        In this test, we check that
        1. The function checks that the blocktemplate's coinbase points to the given Qaddress
        2. The function checks that we are mining on the tip
        :param m_logger:
        :return:
        """
        m_coinbase = Mock(autospec=CoinBase, name='I am a Coinbase')
        m_coinbase.coinbase.addr_to = self.m_mining_address

        m_parent_block = Mock(autospec=Block, name='mock parent_block')
        m_parent_block.transactions = [m_coinbase]

        m_mining_block = Mock(autospec=Block, name='mock _mining_block')
        m_mining_block.transactions = [m_coinbase]
        m_mining_block.mining_blob = b'this is the blob you should iterate the nonce upon'

        self.miner._mining_block = m_mining_block
        self.miner._current_difficulty = StringToUInt256('1')

        # If the coinbase doesn't point to us, make it point to us.
        foreign_qaddress = bob.qaddress
        m_parent_block.headerhash = b'block_headerhash'
        m_mining_block.prev_headerhash = b'block_headerhash'
        blob, difficulty = self.miner.get_block_to_mine(foreign_qaddress.encode(), self.txpool, m_parent_block,
                                                        self.parent_difficulty)

        # actually, the blob's value will not change because mining_block.update_mining_address() is a mock.
        # it will have the same value as in test_get_block_to_mine_perfect_block_no_changes()
        # it's enough to see that it actually runs
        m_mining_block.update_mining_address.assert_called_once()
        self.assertIsNotNone(blob)
        self.assertEqual(difficulty, 1)

    def test_get_block_to_mine_chokes_on_invalid_mining_address(self, m_getTime, m_logger):
        invalid_address = self.m_mining_qaddress + 'aaaa'
        m_parent_block = Mock(autospec=Block, name='mock parent_block')
        with self.assertRaises(ValueError):
            self.miner.get_block_to_mine(invalid_address.encode(), self.txpool, m_parent_block, self.parent_difficulty)

    def test_submit_mined_block(self, m_getTime, m_logger):
        """
        This runs when a miner submits a blob with a valid nonce. It returns True only if
        BlockHeader says the nonce position is okay, and the PoWValidator says the nonce is valid.
        :param m_getTime:
        :param m_logger:
        :return:
        """
        m_mining_block = Mock(autospec=Block, name='mock _mining_block')
        m_mining_block.verify_blob.return_value = False
        self.miner._mining_block = m_mining_block
        blob = 'this is a blob12345that was the nonce'.encode()

        result = self.miner.submit_mined_block(blob)
        self.assertFalse(result)

        m_mining_block.verify_blob.return_value = True
        self.chain_manager.validate_mining_nonce = MagicMock(return_value=False)
        result = self.miner.submit_mined_block(blob)
        self.assertFalse(result)

        m_mining_block.verify_blob.return_value = True
        self.chain_manager.validate_mining_nonce = MagicMock(return_value=True)
        self.m_pre_block_logic.return_value = True
        result = self.miner.submit_mined_block(blob)
        self.assertTrue(result)


@patch('qrl.core.Miner.logger')
@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestMinerWithRealTransactionPool(TestCase):
    def setUp(self):
        self.m_mining_qaddress = alice.qaddress
        self.m_mining_address = parse_qaddress(self.m_mining_qaddress)

        self.chain_manager = Mock(spec=ChainManager)
        self.chain_manager.get_block_size_limit.return_value = 500
        self.chain_manager.get_address_state.return_value = Mock(autospec=AddressState, name='mock alice AddressState')

        self.parent_block = Block()
        self.parent_difficulty = StringToUInt256('0')  # tuple (0,0,0,0,0...) length 32

        self.m_pre_block_logic = Mock(spec=POW.pre_block_logic, name='hello')
        m_add_unprocessed_txn_fn = create_autospec(P2PFactory.add_unprocessed_txn)
        mining_thread_count = 1

        self.miner = Miner(self.m_pre_block_logic,
                           self.m_mining_address,
                           self.chain_manager,
                           mining_thread_count,
                           m_add_unprocessed_txn_fn)

        self.txpool = TransactionPool(None)

        def replacement_set_affected_address(addresses_set):
            return addresses_set.add(alice.address)

        self.m_tx_args = {"addr_from": alice.address,
                          "addrs_to": [bob.address],
                          "amounts": [10],
                          "fee": 1,
                          "PK": alice.pk,
                          "master_addr": None,
                          "size": 150,
                          "validate_extended.return_value": True,
                          "set_affected_address": replacement_set_affected_address
                          }

    @patch('qrl.core.Miner.Block.create')
    def test_create_block_with_one_transaction(self, m_create, m_logger):
        m_tx = Mock(autospec=TransferTransaction, name='mock TransferTransaction')
        m_tx.configure_mock(**self.m_tx_args)

        m_create.return_value = Mock(autospec=Block, name='mock Block', size=205)

        self.txpool.add_tx_to_pool(m_tx, 1, replacement_getTime())

        self.miner.create_block(last_block=self.parent_block, mining_nonce=0, tx_pool=self.txpool,
                                miner_address=self.m_mining_address)

        m_create.assert_called_with(block_number=1, prev_headerhash=b'', prev_timestamp=0, transactions=[m_tx],
                                    miner_address=alice.address)

    @patch('qrl.core.Miner.Block.create')
    def test_create_block_does_not_include_invalid_txs_from_txpool(self, m_create, m_logger):
        self.m_tx_args["validate_extended.return_value"] = False
        m_tx = Mock(autospec=TransferTransaction, name='mock TransferTransaction')
        m_tx.configure_mock(**self.m_tx_args)

        m_create.return_value = Mock(autospec=Block, name='mock Block', size=205)

        self.txpool.add_tx_to_pool(m_tx, 1, replacement_getTime())

        self.miner.create_block(last_block=self.parent_block, mining_nonce=0, tx_pool=self.txpool,
                                miner_address=self.m_mining_address)

        m_create.assert_called_with(block_number=1, prev_headerhash=b'', prev_timestamp=0, transactions=[],
                                    miner_address=alice.address)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase
from mock import patch, Mock, PropertyMock
from pyqrllib.pyqrllib import sha2_256
from collections import OrderedDict

from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss
from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.core.Block import Block
from qrl.core.BlockHeader import BlockHeader
from qrl.core.txs.Transaction import Transaction
from qrl.core.txs.TransferTransaction import TransferTransaction
from qrl.core.txs.CoinBase import CoinBase
from qrl.crypto.misc import merkle_tx_hash

from tests.misc.helper import replacement_getTime

alice = get_alice_xmss()
bob = get_bob_xmss()
slave = get_slave_xmss()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestBlockReal(TestCase):
    # These tests rely on a real BlockHeader
    def setUp(self):
        self.block = Block.create(block_number=5,
                                  prev_headerhash=bytes(sha2_256(b'test')),
                                  prev_timestamp=10,
                                  transactions=[],
                                  miner_address=alice.address)

    def test_update_mining_address(self):
        self.block.update_mining_address(mining_address=bob.address)
        coinbase_tx = Transaction.from_pbdata(self.block.transactions[0])
        self.assertTrue(isinstance(coinbase_tx, CoinBase))
        self.assertEqual(coinbase_tx.addr_to, bob.address)
        hashedtransactions = []
        for tx in self.block.transactions:
            hashedtransactions.append(tx.transaction_hash)
        self.assertEqual(self.block.blockheader.tx_merkle_root, merkle_tx_hash(hashedtransactions))

    def test_mining_blob(self):
        self.block.set_nonces(mining_nonce=5, extra_nonce=4)

        mining_blob = self.block.mining_blob
        self.assertEqual(len(mining_blob), config.dev.mining_blob_size)
        mining_nonce_bytes = mining_blob[config.dev.mining_nonce_offset:config.dev.mining_nonce_offset + 4]
        extra_nonce_bytes = mining_blob[config.dev.extra_nonce_offset:config.dev.extra_nonce_offset + 8]

        mining_nonce = int.from_bytes(mining_nonce_bytes, byteorder='big', signed=False)
        extra_nonce = int.from_bytes(extra_nonce_bytes, byteorder='big', signed=False)

        self.assertEqual(mining_nonce, 5)
        self.assertEqual(extra_nonce, 4)

    def test_serialize_deserialize(self):
        output = self.block.serialize()
        block_2 = Block.deserialize(output)
        self.assertEqual(block_2, self.block)


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestBlock(TestCase):
    def setUp(self):
        self.blockheader = Mock(name='mock BlockHeader', autospec=BlockHeader, block_number=5,
                                headerhash=bytes(sha2_256(b'mock headerhash')),
                                prev_headerhash=bytes(sha2_256(b'test')))

        self.block = Block.create(block_number=5,
                                  prev_headerhash=bytes(sha2_256(b'test')),
                                  prev_timestamp=10,
                                  transactions=[],
                                  miner_address=alice.address)
        self.block.blockheader = self.blockheader

    def test_set_mining_nonce_from_blob(self):
        current_mining_nonce = self.block.mining_nonce
        current_headerhash = self.block.headerhash
        mining_blob = self.block.mining_blob
        self.block.blockheader.set_mining_nonce_from_blob(mining_blob)
        self.assertEqual(self.block.blockheader.mining_nonce, current_mining_nonce)
        self.assertEqual(self.block.headerhash, current_headerhash)
        self.assertEqual(self.block.blockheader.mining_blob, mining_blob)

    def test_epoch(self):
        # Epoch should follow this formula: block_number // blocks_per_epoch
        self.assertEqual(self.block.epoch, int(self.block.block_number // config.dev.blocks_per_epoch))

    def test_verify_blob(self):
        m_blockheader = Mock()
        self.block.blockheader = m_blockheader

        self.block.verify_blob(b'blob')
        m_blockheader.verify_blob.assert_called_once_with(b'blob')

    def test_validate(self):
        # 1. Set up all the mocking so that Block.validate() should pass
        m_chain_manager = Mock(name='Mock ChainManager')
        attrs_all_pass = {
            'get_block_is_duplicate.return_value': False,
            'validate_mining_nonce.return_value': True,
        }
        m_chain_manager.configure_mock(**attrs_all_pass)
        self.block._validate_parent_child_relation = Mock(return_value=True)

        result = self.block.validate(m_chain_manager, OrderedDict())
        self.assertTrue(result)

        # 2. Switch the mock checks one by one to invalid, and make sure that validate() returns False
        # The Block is already in the State (a duplicate)
        m_chain_manager.get_block_is_duplicate.return_value = True
        result = self.block.validate(m_chain_manager, OrderedDict())
        self.assertFalse(result)
        m_chain_manager.get_block_is_duplicate.return_value = False

        # The mining nonce is invalid
        m_chain_manager.validate_mining_nonce.return_value = False
        result = self.block.validate(m_chain_manager, OrderedDict())
        self.assertFalse(result)
        m_chain_manager.validate_mining_nonce.return_value = True

        # No parent block found, and it's not in future_blocks either
        m_chain_manager.get_block.return_value = None
        result = self.block.validate(m_chain_manager, OrderedDict())
        self.assertFalse(result)
        m_chain_manager.get_block.return_value = Mock(name='mock Block')

        # The parent_block is not actually Block's parent
        self.block._validate_parent_child_relation.return_value = False
        result = self.block.validate(m_chain_manager, OrderedDict())
        self.assertFalse(result)
        self.block._validate_parent_child_relation.return_value = True

        # Block.transactions is [] (it should at least have the CoinBase TX in there)
        with patch('qrl.core.Block.Block.transactions', new_callable=PropertyMock) as m_transactions:
            m_transactions.return_value = []
            result = self.block.validate(m_chain_manager, OrderedDict())
            self.assertFalse(result)

        # There was a problem with the CoinBase TX
        with patch('qrl.core.txs.CoinBase.CoinBase.validate_extended') as m_validate_extended:
            m_validate_extended.return_value = False
            result = self.block.validate(m_chain_manager, OrderedDict())
            self.assertFalse(result)
            m_validate_extended.return_value = None

            m_validate_extended.side_effect = Exception
            result = self.block.validate(m_chain_manager, OrderedDict())
            self.assertFalse(result)

        # The BlockHeader doesn't fit with this Block
        self.blockheader.validate.return_value = False
        result = self.block.validate(m_chain_manager, OrderedDict())
        self.assertFalse(result)

    def test_is_future_block(self):
        self.blockheader.timestamp = replacement_getTime() + config.dev.block_max_drift + 1
        result = self.block.is_future_block()
        self.assertTrue(result)

        self.blockheader.timestamp = replacement_getTime()
        result = self.block.is_future_block()
        self.assertFalse(result)

    def test_validate_parent_child_relation(self):
        self.block._validate_parent_child_relation(Mock())
        self.blockheader.validate_parent_child_relation.assert_called_once()


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
@patch('qrl.core.txs.CoinBase.CoinBase.apply_state_changes', name='mock CoinBase.apply_state_changes()',
       return_value=True)
@patch('qrl.core.txs.TransferTransaction.TransferTransaction.apply_state_changes',
       name='mock Transfer.apply_state_changes()', return_value=True)
@patch('qrl.core.txs.TransferTransaction.TransferTransaction.validate_extended',
       name='mock Transfer.validate_extended()', return_value=True)
@patch('qrl.core.txs.TransferTransaction.TransferTransaction.validate', name='mock Transfer.validate()',
       return_value=True)
class TestBlockApplyStateChanges(TestCase):
    """
    Block.apply_state_changes() performs validation on the Transactions again before calling their apply_state_changes()

    If I use Mock Transactions, I need to patch out Block's copying of TX.pbdata into Block's own stuff.
    Then patch out the reconstruction of the Transaction.from_pbdata() in apply_state_changes().
    What value does this give me?
    Not having to individually patch out all Transaction methods used in apply_state_changes() (thus making me reliant
    on apply_state_changes() implementation).
    But to make the test fail/pass, I still need to set certain functions to pass/fail, which makes the test somewhat
    reliant on the implementation.

    If I use real Transactions, I end up testing more of the code, and I am surer that it works well.
    I still have to patch out certain functions to set them to pass/fail, so I am still somewhat reliant on the
    implementation.
    """

    def generate_address_states(self, alice_attrs, bob_attrs, slave_attrs):
        address_states = {
            self.alice.address: Mock(name='self.alice AddressState', autospec=AddressState, **alice_attrs),
            self.bob.address: Mock(name='self.bob AddressState', autospec=AddressState, **bob_attrs),
            self.slave.address: Mock(name='self.slave AddressState', autospec=AddressState, **slave_attrs)
        }
        return address_states

    def setUp(self):
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()
        self.slave = get_slave_xmss()

        self.tx1 = TransferTransaction.create(addrs_to=[self.bob.address], amounts=[10], fee=1, xmss_pk=self.alice.pk)
        self.tx2 = TransferTransaction.create(addrs_to=[self.bob.address], amounts=[10], fee=1, xmss_pk=self.slave.pk,
                                              master_addr=self.alice.address)
        # self.tx1 = Mock(autospec=TransferTransaction, addrs_to=[self.bob.address], amounts=[10], fee=1,
        #                 xmss_pk=self.alice.pk)
        # self.tx2 = Mock(autospec=TransferTransaction, addrs_to=[self.bob.address], amounts=[10], fee=1,
        #                 xmss_pk=self.slave.pk, master_addr=self.alice.address)
        self.tx1._data.nonce = 3
        self.tx2._data.nonce = 6
        self.tx1.sign(self.alice)
        self.tx2.sign(self.slave)

        self.block_attrs = {
            "block_number": 5,
            "prev_headerhash": bytes(sha2_256(b'test')),
            "prev_timestamp": 10,
            "transactions": [self.tx1, self.tx2],
            "miner_address": self.alice.address
        }
        self.alice_addrstate_attrs = {'nonce': 2, 'ots_key_reuse.return_value': False}
        self.slave_addrstate_attrs = {'nonce': 5, 'ots_key_reuse.return_value': False}

    def test_all_ok(self, m_TransferTransaction_validate, m_TransferTransaction_validate_extended,
                    m_TransferTransaction_apply_state_changes, m_CoinBase_apply_state_changes):
        address_states = self.generate_address_states(self.alice_addrstate_attrs, {}, self.slave_addrstate_attrs)

        block = Block.create(**self.block_attrs)
        result = block.apply_state_changes(address_states)
        self.assertTrue(result)

    def test_extra_coinbase_tx(self, m_TransferTransaction_validate,
                               m_TransferTransaction_validate_extended,
                               m_TransferTransaction_apply_state_changes,
                               m_CoinBase_apply_state_changes):
        address_states = self.generate_address_states(self.alice_addrstate_attrs, {}, self.slave_addrstate_attrs)

        coinbase_extra = CoinBase.create(500, self.alice.address, 5)
        self.block_attrs["transactions"] = [self.tx1, coinbase_extra, self.tx2]

        block = Block.create(**self.block_attrs)
        result = block.apply_state_changes(address_states)
        self.assertFalse(result)

    def test_bad_nonce_or_ots_reused(self, m_TransferTransaction_validate,
                                     m_TransferTransaction_validate_extended,
                                     m_TransferTransaction_apply_state_changes,
                                     m_CoinBase_apply_state_changes):
        # If a TX was signed by a Slave XMSS, apply_state_changes() should check against the Slave's AddressState.nonce.
        # In this case, tx.nonce = 3 but slave addrstate.nonce = 0
        self.slave_addrstate_attrs['nonce'] = 0
        address_states = self.generate_address_states(self.alice_addrstate_attrs, {}, self.slave_addrstate_attrs)

        block = Block.create(**self.block_attrs)
        result = block.apply_state_changes(address_states)
        self.assertFalse(result)
        self.slave_addrstate_attrs['nonce'] = 5

        # Now we pretend that Alice's OTS key has been reused.
        self.alice_addrstate_attrs['ots_key_reuse.return_value'] = True
        address_states = self.generate_address_states(self.alice_addrstate_attrs, {}, self.slave_addrstate_attrs)
        result = block.apply_state_changes(address_states)
        self.assertFalse(result)
        self.alice_addrstate_attrs['ots_key_reuse.return_value'] = False

        # Now we pretend that Slave's OTS key has been reused.
        self.slave_addrstate_attrs['ots_key_reuse.return_value'] = True
        address_states = self.generate_address_states(self.alice_addrstate_attrs, {}, self.slave_addrstate_attrs)
        result = block.apply_state_changes(address_states)
        self.assertFalse(result)

    def test_tx_validation_fails(self, m_TransferTransaction_validate,
                                 m_TransferTransaction_validate_extended,
                                 m_TransferTransaction_apply_state_changes,
                                 m_CoinBase_apply_state_changes):
        address_states = self.generate_address_states(self.alice_addrstate_attrs, {}, self.slave_addrstate_attrs)
        block = Block.create(**self.block_attrs)

        m_TransferTransaction_validate.return_value = False
        result = block.apply_state_changes(address_states)
        self.assertFalse(result)
        m_TransferTransaction_validate.return_value = True

        m_TransferTransaction_validate_extended.return_value = False
        result = block.apply_state_changes(address_states)
        self.assertFalse(result)
        m_TransferTransaction_validate_extended.return_value = True

        with patch('qrl.core.txs.CoinBase.CoinBase.validate_extended') as m_validate_extended:
            m_validate_extended.return_value = False
            result = block.apply_state_changes(address_states)
            self.assertFalse(result)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from google.protobuf.json_format import MessageToJson, Parse

from qrl.core import config
from qrl.generated import qrl_pb2


class BlockMetadata(object):

    def __init__(self, pbdata=None):
        self._data = pbdata
        if not pbdata:
            self._data = qrl_pb2.BlockMetaData()
            self._data.block_difficulty = bytes([0] * 32)
            self._data.cumulative_difficulty = bytes([0] * 32)
        else:
            # TODO: Improve validation
            if len(self.cumulative_difficulty) != 32:
                raise ValueError("Invalid cumulative_difficulty")

            if len(self.block_difficulty) != 32:
                raise ValueError("Invalid block_difficulty")

    @property
    def pbdata(self):
        return self._data

    @property
    def block_difficulty(self):
        return tuple(self._data.block_difficulty)

    @property
    def cumulative_difficulty(self):
        return tuple(self._data.cumulative_difficulty)

    @property
    def child_headerhashes(self):
        return self._data.child_headerhashes

    @property
    def last_N_headerhashes(self):
        return self._data.last_N_headerhashes

    def set_block_difficulty(self, value):

        if len(value) != 32:
            raise ValueError("Invalid block_difficulty")

        self._data.block_difficulty = bytes(value)

    def set_cumulative_difficulty(self, value):
        if len(value) != 32:
            raise ValueError("Invalid cumulative_difficulty")
        self._data.cumulative_difficulty = bytes(value)

    def add_child_headerhash(self, child_headerhash: bytes):
        if child_headerhash not in self._data.child_headerhashes:
            self._data.child_headerhashes.append(child_headerhash)

    def update_last_headerhashes(self, parent_last_N_headerhashes, last_headerhash: bytes):
        del self._data.last_N_headerhashes[:]
        self._data.last_N_headerhashes.extend(parent_last_N_headerhashes)
        self._data.last_N_headerhashes.append(last_headerhash)
        if len(self._data.last_N_headerhashes) > config.dev.N_measurement:
            del self._data.last_N_headerhashes[0]

        if len(self._data.last_N_headerhashes) > config.dev.N_measurement:
            raise Exception('Size of last_N_headerhashes is more than expected %s %s',
                            len(self._data.last_N_headerhashes),
                            config.dev.N_measurement)

    @staticmethod
    def create(block_difficulty=bytes([0] * 32),
               cumulative_difficulty=bytes([0] * 32),
               child_headerhashes=None):
        block_meta_data = BlockMetadata()
        block_meta_data._data.block_difficulty = block_difficulty
        block_meta_data._data.cumulative_difficulty = cumulative_difficulty

        if child_headerhashes:
            for headerhash in child_headerhashes:
                block_meta_data._data.child_headerhashes.append(headerhash)

        return block_meta_data

    @staticmethod
    def from_json(json_data):
        pbdata = qrl_pb2.BlockMetaData()
        Parse(json_data, pbdata)
        return BlockMetadata(pbdata)

    def to_json(self) -> str:
        return MessageToJson(self._data, sort_keys=True).encode()

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qrl_pb2.BlockMetaData()
        pbdata.ParseFromString(bytes(data))
        return BlockMetadata(pbdata)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import os
from configparser import RawConfigParser
from os.path import pardir
from unittest import TestCase

from qrl.core.misc import logger

logger.initialize_default()


class TestHelpers(TestCase):
    THIS_DIR = os.path.dirname(os.path.realpath(__file__))
    PATH_SETUP_CFG = os.path.join(THIS_DIR, pardir, pardir, "setup.cfg")
    PATH_REQUIREMENTS = os.path.join(THIS_DIR, pardir, pardir, "requirements.txt")

    def __init__(self, *args, **kwargs):
        super(TestHelpers, self).__init__(*args, **kwargs)

    def test_setup_requirements(self):
        config = RawConfigParser()
        config.read(self.PATH_SETUP_CFG)

        install_requires_in_file = config.get('options', 'install_requires')
        install_requires = set()
        for item in install_requires_in_file.split():
            install_requires.add(item.strip())

        with open(self.PATH_REQUIREMENTS) as f:
            all_req = f.readlines()
            for r in all_req:
                r = r.strip()
                if r.startswith('#'):
                    continue

                if r not in install_requires:
                    raise Exception("{} not found in setup.cfg".format(r))

                install_requires.remove(r)

        if len(install_requires) > 0:
            raise Exception("Not found in requirements.txt:\n{}".format("\n".join(install_requires)))



from pyqrllib.pyqrllib import bin2hstr

from qrl.core import config
from qrl.core.AddressState import AddressState
from qrl.core.misc import logger
from qrl.core.txs.Transaction import Transaction


class SlaveTransaction(Transaction):

    def __init__(self, protobuf_transaction=None):
        super(SlaveTransaction, self).__init__(protobuf_transaction)

    @property
    def slave_pks(self):
        return self._data.slave.slave_pks

    @property
    def access_types(self):
        return self._data.slave.access_types

    def get_data_bytes(self) -> bytes:
        tmptxhash = (self.master_addr +
                     self.fee.to_bytes(8, byteorder='big', signed=False))

        for index in range(0, len(self.slave_pks)):
            tmptxhash = (tmptxhash +
                         self.slave_pks[index] +
                         self.access_types[index].to_bytes(8, byteorder='big', signed=False))

        return tmptxhash

    @staticmethod
    def create(slave_pks: list, access_types: list, fee: int, xmss_pk: bytes, master_addr: bytes = None):
        transaction = SlaveTransaction()

        if master_addr:
            transaction._data.master_addr = master_addr

        for slave_pk in slave_pks:
            transaction._data.slave.slave_pks.append(slave_pk)
        for access_type in access_types:
            transaction._data.slave.access_types.append(access_type)
        transaction._data.fee = fee

        transaction._data.public_key = xmss_pk

        transaction.validate_or_raise(verify_signature=False)

        return transaction

    def _validate_custom(self) -> bool:
        if (len(self.slave_pks) > config.dev.transaction_multi_output_limit or
                len(self.access_types) > config.dev.transaction_multi_output_limit):
            logger.warning('List has more than 100 slave pks or access_types')
            logger.warning('Slave pks len %s', len(self.slave_pks))
            logger.warning('Access types len %s', len(self.access_types))
            return False

        if len(self.slave_pks) != len(self.access_types):
            logger.warning('Number of slave pks are not equal to the number of access types provided')
            logger.warning('Slave pks len %s', len(self.slave_pks))
            logger.warning('Access types len %s', len(self.access_types))
            return False

        for access_type in self.access_types:
            if access_type not in [0, 1]:
                logger.warning('Invalid Access type %s', access_type)
                return False

        return True

    def validate_extended(self, addr_from_state: AddressState, addr_from_pk_state: AddressState) -> bool:
        if not self.validate_slave(addr_from_state, addr_from_pk_state):
            return False

        tx_balance = addr_from_state.balance

        if self.fee < 0:
            logger.info('Slave: State validation failed for %s because: Negative send', bin2hstr(self.txhash))
            return False

        if tx_balance < self.fee:
            logger.info('Slave: State validation failed for %s because: Insufficient funds', bin2hstr(self.txhash))
            logger.info('balance: %s, amount: %s', tx_balance, self.fee)
            return False

        if addr_from_pk_state.ots_key_reuse(self.ots_key):
            logger.info('Slave: State validation failed for %s because: OTS Public key re-use detected',
                        bin2hstr(self.txhash))
            return False

        return True

    def apply_state_changes(self, addresses_state):
        if self.addr_from in addresses_state:
            addresses_state[self.addr_from].balance -= self.fee
            for index in range(0, len(self.slave_pks)):
                addresses_state[self.addr_from].add_slave_pks_access_type(self.slave_pks[index],
                                                                          self.access_types[index])
            addresses_state[self.addr_from].transaction_hashes.append(self.txhash)

        self._apply_state_changes_for_PK(addresses_state)

    def revert_state_changes(self, addresses_state, chain_manager):
        if self.addr_from in addresses_state:
            addresses_state[self.addr_from].balance += self.fee
            for index in range(0, len(self.slave_pks)):
                addresses_state[self.addr_from].remove_slave_pks_access_type(self.slave_pks[index])
            addresses_state[self.addr_from].transaction_hashes.remove(self.txhash)

        self._revert_state_changes_for_PK(addresses_state, chain_manager)



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.

from google.protobuf.json_format import MessageToJson, Parse

from qrl.generated import qrl_pb2


class TokenMetadata(object):
    def __init__(self, protobuf_data=None):
        self._data = protobuf_data
        if protobuf_data is None:
            self._data = qrl_pb2.TokenMetadata()

    @property
    def pbdata(self):
        """
        Returns a protobuf object that contains persistable data representing this object
        :return: A protobuf TokenMetadata object
        :rtype: qrl_pb2.TokenMetadata
        """
        return self._data

    @property
    def token_txhash(self):
        return self._data.token_txhash

    @property
    def transfer_token_tx_hashes(self):
        return self._data.transfer_token_tx_hashes

    @staticmethod
    def create(token_txhash: bytes, transfer_token_txhashes: list):
        token_metadata = TokenMetadata()

        token_metadata._data.token_txhash = token_txhash

        token_metadata.update(transfer_token_txhashes)

        return token_metadata

    def update(self, transfer_token_txhashes: list):
        for transfer_token_txhash in transfer_token_txhashes:
            self._data.transfer_token_tx_hashes.extend([transfer_token_txhash])

    def remove(self, transfer_token_txhash: bytes):
        i = 0
        while i < len(self._data.transfer_token_tx_hashes):
            if self._data.transfer_token_tx_hashes[i] == transfer_token_txhash:
                del self._data.transfer_token_tx_hashes[i]
                return
            i += 1

    def to_json(self):
        return MessageToJson(self._data, sort_keys=True)

    @staticmethod
    def from_json(json_data):
        pbdata = qrl_pb2.TokenMetadata()
        Parse(json_data, pbdata)
        return TokenMetadata(pbdata)

    def serialize(self) -> str:
        return self._data.SerializeToString()

    @staticmethod
    def deserialize(data):
        pbdata = qrl_pb2.TokenMetadata()
        pbdata.ParseFromString(bytes(data))
        return TokenMetadata(pbdata)



from unittest import TestCase

from mock import patch, Mock
from pyqrllib.pyqrllib import bin2hstr

from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.ChainManager import ChainManager
from qrl.core.txs.TokenTransaction import TokenTransaction
from qrl.generated import qrl_pb2
from tests.misc.helper import get_alice_xmss, get_bob_xmss, get_slave_xmss

logger.initialize_default()


@patch('qrl.core.txs.Transaction.logger')
class TestTokenTransactionStateChanges(TestCase):
    def setUp(self):
        self.alice = get_alice_xmss()
        self.bob = get_bob_xmss()

        self.params = {
            "symbol": b'QRL',
            "name": b'Quantum Resistant Ledger',
            "owner": self.alice.address,
            "decimals": 15,
            "initial_balances": [],
            "fee": 1,
            "xmss_pk": self.alice.pk
        }

        self.unused_chain_manager_mock = Mock(autospec=ChainManager, name='unused ChainManager')

    def generate_addresses_state(self, tx):
        addresses_state = {
            self.alice.address: Mock(autospec=AddressState, name='alice AddressState',
                                     tokens={bin2hstr(tx.txhash): 0}, transaction_hashes=[],
                                     balance=100),
            self.bob.address: Mock(autospec=AddressState, name='bob AddressState', tokens={bin2hstr(tx.txhash): 0},
                                   transaction_hashes=[],
                                   balance=0),
        }
        return addresses_state

    def test_apply_state_changes(self, m_logger):
        """
        Alice creates a token. Obviously, she gives herself some of this token.
        But she also gives Bob some tokens too.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = self.generate_addresses_state(tx)

        # According to the State, Alice has 100 coins, and Bob has 0 coins.
        # After applying the Transaction, Alice and Bob should have 1000 tokens, and Alice's balance should be 99.
        # AddressState.transaction_hashes now also reference the TokenTransaction that created the Tokens.
        tx.apply_state_changes(addresses_state)
        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(tx.txhash, 1000)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(tx.txhash, 1000)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[self.bob.address].transaction_hashes)
        addresses_state[self.alice.address].increase_nonce.assert_called_once()
        addresses_state[self.alice.address].set_ots_key.assert_called_once()
        addresses_state[self.bob.address].increase_nonce.assert_not_called()
        addresses_state[self.bob.address].set_ots_key.assert_not_called()

    def test_apply_state_changes_empty_addresses_state(self, m_logger):
        """
        After applying the Transaction, Alice and Bob should have 1000 tokens, and Alice's balance should be 99.
        AddressState.transaction_hashes now also reference the TokenTransaction that created the Tokens.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances
        addresses_state_empty = {}

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)

        tx.apply_state_changes(addresses_state_empty)
        self.assertEqual(addresses_state_empty, {})

    def test_apply_state_changes_owner_not_in_address_state(self, m_logger):
        """
        In this case, Alice didn't give herself any tokens. How generous! She gave them all to Bob.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)

        # Signing the TX also generates the txhash, which we need to generate the AddressState properly.
        addresses_state = self.generate_addresses_state(tx)
        tx.apply_state_changes(addresses_state)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[self.bob.address].transaction_hashes)
        addresses_state[self.alice.address].increase_nonce.assert_called_once()
        addresses_state[self.alice.address].set_ots_key.assert_called_once()
        addresses_state[self.bob.address].increase_nonce.assert_not_called()
        addresses_state[self.bob.address].set_ots_key.assert_not_called()

    def test_apply_state_changes_signed_by_slave_xmss(self, m_logger):
        """
        Alice creates a token, gives herself and Bob some tokens.
        But she uses a XMSS slave to sign it.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        slave = get_slave_xmss()
        self.params["initial_balances"] = initial_balances
        self.params["xmss_pk"] = slave.pk
        self.params["master_addr"] = self.alice.address
        tx = TokenTransaction.create(**self.params)
        tx.sign(slave)

        # Now that we have the Slave XMSS address, we should add it to AddressState so that apply_state_changes()
        # can do something with it
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[slave.address] = Mock(autospec=AddressState,
                                              name='slave AddressState',
                                              tokens={bin2hstr(tx.txhash): 0},
                                              transaction_hashes=[],
                                              balance=0)

        tx.apply_state_changes(addresses_state)

        self.assertEqual(addresses_state[self.alice.address].balance, 99)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(tx.txhash, 1000)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(tx.txhash, 1000)
        self.assertEqual([tx.txhash], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[slave.address].transaction_hashes)
        self.assertEqual([tx.txhash], addresses_state[self.bob.address].transaction_hashes)
        addresses_state[slave.address].increase_nonce.assert_called_once()
        addresses_state[slave.address].set_ots_key.assert_called_once()
        addresses_state[self.alice.address].increase_nonce.assert_not_called()
        addresses_state[self.alice.address].set_ots_key.assert_not_called()

    def test_revert_state_changes(self, m_logger):
        """
        Same setup as in test_apply_state_changes(). This time though, the changes have already been applied,
        and we would like to roll them back.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)
        # Apply the changes!
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].tokens[bin2hstr(tx.txhash)] = 1000
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]
        addresses_state[self.bob.address].tokens[bin2hstr(tx.txhash)] = 1000
        addresses_state[self.bob.address].transaction_hashes = [tx.txhash]

        # After applying the Transaction, it should be as if Alice had never created the tokens in the first place.
        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(tx.txhash, -1000)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(tx.txhash, -1000)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)
        addresses_state[self.alice.address].decrease_nonce.assert_called_once()
        addresses_state[self.alice.address].unset_ots_key.assert_called_once()
        addresses_state[self.bob.address].decrease_nonce.assert_not_called()
        addresses_state[self.bob.address].unset_ots_key.assert_not_called()

    def test_revert_state_changes_empty_addresses_state(self, m_logger):
        """If we didn't have any AddressStates for the addresses involved in this test, do nothing"""
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)
        addresses_state = {}

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state, {})

    def test_revert_state_changes_owner_not_in_address_state(self, m_logger):
        """
        In this case, Alice didn't give herself any tokens. How generous! She gave them all to Bob.
        But we want to revert this.
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        self.params["initial_balances"] = initial_balances

        tx = TokenTransaction.create(**self.params)
        tx.sign(self.alice)

        addresses_state = self.generate_addresses_state(tx)
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]
        addresses_state[self.bob.address].tokens[bin2hstr(tx.txhash)] = 1000
        addresses_state[self.bob.address].transaction_hashes = [tx.txhash]

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        self.assertEqual(addresses_state[self.bob.address].balance, 0)
        addresses_state[self.alice.address].update_token_balance.assert_not_called()
        addresses_state[self.bob.address].update_token_balance.assert_called_with(tx.txhash, -1000)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)
        addresses_state[self.alice.address].decrease_nonce.assert_called_once()
        addresses_state[self.alice.address].unset_ots_key.assert_called_once()
        addresses_state[self.bob.address].decrease_nonce.assert_not_called()
        addresses_state[self.bob.address].unset_ots_key.assert_not_called()

    def test_revert_state_changes_signed_by_slave_xmss(self, m_logger):
        """
        Alice creates a token, gives herself and Bob some tokens.
        But she uses a XMSS slave to sign it.
        Can we undo it?
        """
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        slave = get_slave_xmss()
        self.params["initial_balances"] = initial_balances
        self.params["xmss_pk"] = slave.pk
        self.params["master_addr"] = self.alice.address
        tx = TokenTransaction.create(**self.params)
        tx.sign(slave)

        # Now that we have the Slave XMSS address, we should add it to AddressState so that apply_state_changes()
        # can do something with it
        addresses_state = self.generate_addresses_state(tx)
        addresses_state[slave.address] = Mock(autospec=AddressState,
                                              name='slave AddressState',
                                              tokens={bin2hstr(tx.txhash): 0},
                                              transaction_hashes=[],
                                              balance=0)
        # Also, update the AddressStates manually!
        addresses_state[self.alice.address].balance = 99
        addresses_state[self.alice.address].transaction_hashes = [tx.txhash]
        addresses_state[self.alice.address].tokens[bin2hstr(tx.txhash)] = 1000
        addresses_state[self.bob.address].balance = 0
        addresses_state[self.bob.address].transaction_hashes = [tx.txhash]
        addresses_state[self.bob.address].tokens[bin2hstr(tx.txhash)] = 1000
        addresses_state[slave.address].transaction_hashes = [tx.txhash]

        tx.revert_state_changes(addresses_state, self.unused_chain_manager_mock)

        self.assertEqual(addresses_state[self.alice.address].balance, 100)
        addresses_state[self.alice.address].update_token_balance.assert_called_with(tx.txhash, -1000)
        addresses_state[self.bob.address].update_token_balance.assert_called_with(tx.txhash, -1000)
        self.assertEqual([], addresses_state[self.alice.address].transaction_hashes)
        self.assertEqual([], addresses_state[slave.address].transaction_hashes)
        self.assertEqual([], addresses_state[self.bob.address].transaction_hashes)
        addresses_state[slave.address].decrease_nonce.assert_called_once()
        addresses_state[slave.address].unset_ots_key.assert_called_once()
        addresses_state[self.alice.address].decrease_nonce.assert_not_called()
        addresses_state[self.alice.address].unset_ots_key.assert_not_called()

    def test_validate_tx(self, m_logger):
        initial_balances = [qrl_pb2.AddressAmount(address=self.alice.address, amount=1000),
                            qrl_pb2.AddressAmount(address=self.bob.address, amount=1000)]
        slave = get_slave_xmss()
        self.params["initial_balances"] = initial_balances
        self.params["xmss_pk"] = slave.pk
        self.params["master_addr"] = self.alice.address
        tx = TokenTransaction.create(**self.params)
        tx.sign(slave)

        self.assertTrue(tx.validate_or_raise())

        tx._data.transaction_hash = b'abc'

        # Should fail, as we have modified with invalid transaction_hash
        with self.assertRaises(ValueError):
            tx.validate_or_raise()



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
import contextlib
from unittest import TestCase
from mock import Mock, MagicMock, patch

from pyqryptonight.pyqryptonight import StringToUInt256

from tests.misc.helper import get_alice_xmss, get_random_xmss
from qrl.core import config
from qrl.core.misc import logger
from qrl.core.AddressState import AddressState
from qrl.core.ESyncState import ESyncState
from qrl.core.GenesisBlock import GenesisBlock
from qrl.core.node import POW

from tests.misc.helper import replacement_getTime

logger.initialize_default()


@contextlib.contextmanager
def set_mining_enabled(new_value):
    old_value = config.user.mining_enabled
    try:
        config.user.mining_enabled = new_value
        yield
    finally:
        config.user.mining_enabled = old_value


@patch('qrl.core.misc.ntp.getTime', new=replacement_getTime)
class TestNode(TestCase):
    def __init__(self, *args, **kwargs):
        super(TestNode, self).__init__(*args, **kwargs)

    def test_create(self):
        chain_manager = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POW(chain_manager=chain_manager,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider,
                   mining_address=get_random_xmss().address,
                   mining_thread_count=0)

        self.assertIsNotNone(node)

    def test_sync_state_change_unsynced(self):
        chain_manager = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POW(chain_manager=chain_manager,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider,
                   mining_address=get_random_xmss().address,
                   mining_thread_count=0)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.unsynced)
        # FIXME: Add more asserts

    def test_sync_state_change_syncing(self):
        chain_manager = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POW(chain_manager=chain_manager,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider,
                   mining_address=get_random_xmss().address,
                   mining_thread_count=0)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.syncing)
        # FIXME: Add more asserts

    def test_sync_state_change_synced(self):
        chain_manager = Mock()
        chain_manager.height = 0
        chain_manager.get_block = MagicMock(return_value=GenesisBlock())
        chain_manager.last_block = GenesisBlock()
        chain_manager.tx_pool = Mock()
        chain_manager.tx_pool.transaction_pool = []
        chain_manager.tx_pool.transactions = chain_manager.tx_pool.transaction_pool

        get_block_metadata_response = Mock()
        get_block_metadata_response.block_difficulty = StringToUInt256('2')
        chain_manager.get_block_metadata = MagicMock(return_value=get_block_metadata_response)

        alice_xmss = get_alice_xmss()
        chain_manager._state.get_address_state = MagicMock(return_value=AddressState.get_default(alice_xmss.address))
        chain_manager._state.get_measurement = MagicMock(return_value=60)

        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()
        # Setting mining enabled False, when update_note_state set to synced,
        # starts miner which is not exited properly by unit test
        with set_mining_enabled(False):
            node = POW(chain_manager=chain_manager,
                       p2p_factory=p2p_factory,
                       sync_state=sync_state,
                       time_provider=time_provider,
                       mining_address=get_random_xmss().address,
                       mining_thread_count=0)

            self.assertIsNotNone(node)
            node.update_node_state(ESyncState.synced)
        # FIXME: Add more asserts

    def test_sync_state_change_forked(self):
        chain_manager = Mock()
        p2p_factory = Mock()
        sync_state = Mock()
        time_provider = Mock()

        node = POW(chain_manager=chain_manager,
                   p2p_factory=p2p_factory,
                   sync_state=sync_state,
                   time_provider=time_provider,
                   mining_address=get_random_xmss().address,
                   mining_thread_count=0)

        self.assertIsNotNone(node)
        node.update_node_state(ESyncState.forked)
        # FIXME: Add more asserts



# coding=utf-8
# Distributed under the MIT software license, see the accompanying
# file LICENSE or http://www.opensource.org/licenses/mit-license.php.
from unittest import TestCase

from mock import Mock, patch
from pyqryptonight.pyqryptonight import StringToUInt256

from qrl.core.misc import logger
from qrl.core.Block import Block
from qrl.core.messagereceipt import MessageReceipt
from qrl.core.p2p.p2pChainManager import P2PChainManager
from qrl.core.p2p.p2pfactory import P2PFactory
from qrl.core.p2p.p2pprotocol import P2PProtocol
from qrl.crypto.misc import sha256
from qrl.generated import qrl_pb2, qrllegacy_pb2

logger.initialize_default()


def make_message(**kwargs):
    return qrllegacy_pb2.LegacyMessage(**kwargs)


@patch('qrl.core.p2p.p2pChainManager.logger')
class TestP2PChainManager(TestCase):
    def setUp(self):
        self.manager = P2PChainManager()
        self.channel = Mock(autospec=P2PProtocol, addr_remote='1.1.1.1')
        self.channel.factory = Mock(autospec=P2PFactory)
        self.channel.factory.master_mr = Mock(autospec=MessageReceipt)

    def test_new_channel_registers_with_observables(self, m_logger):
        self.manager.new_channel(self.channel)
        self.channel.register.assert_called()

    def test_handle_fetch_block(self, m_logger):
        """
        1. A peer has sent a request for a block.
        2. This function serves the block in response to that request.
        :return:
        """
        self.channel.factory.chain_height = 10
        self.channel.factory.get_block_by_number.return_value = Mock(autospec=Block, pbdata=qrl_pb2.Block())
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.FB, fbData=qrllegacy_pb2.FBData(index=1))
        self.manager.handle_fetch_block(self.channel, msg)

        self.channel.send.assert_called()

    def test_handle_fetch_block_bad_height(self, m_logger):
        """
        If the peer has sent a request for a block which we don't have, nothing should happen.
        :return:
        """
        self.channel.factory.chain_height = 10
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.FB, fbData=qrllegacy_pb2.FBData(index=100))
        self.manager.handle_fetch_block(self.channel, msg)

        self.channel.send.assert_not_called()

    def test_handle_push_block(self, m_logger):
        """
        1. This node has requested a peer to send us its MessageReceipts
        2. The node has received a list of Block headers
        3. The node has requested a specific block.
        4. The peer sends the new block, which is handled by this function?
        :return:
        """
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.PB, pbData=qrllegacy_pb2.PBData(block=qrl_pb2.Block()))
        self.manager.handle_push_block(self.channel, msg)

        self.channel.factory.block_received.assert_called()

    @patch('qrl.core.p2p.p2pChainManager.Block')
    def test_handle_push_block_bad_block(self, m_Block, m_logger):
        """
        If the Block couldn't be constructed from the Protobuf data, nothing should happen.
        (including disconnecting the peer? dunno)
        :return:
        """
        # Case 1: Block() raises an Exception
        m_Block.side_effect = Exception
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.PB, pbData=qrllegacy_pb2.PBData(block=qrl_pb2.Block()))
        self.manager.handle_push_block(self.channel, msg)

        self.channel.factory.block_received.assert_not_called()

        # Case 2: no protobuf data was in the message.
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.PB)
        self.manager.handle_push_block(self.channel, msg)

        self.channel.factory.block_received.assert_not_called()

    def test_handle_block(self, m_logger):
        """
        1. A peer has found a new block. It broadcasts the MessageReceipt for that block.
        2. This node finds that it hasn't got that block yet, so it requests that block.
        3. The peer sends the new block, which is handled by this function.
        :return:
        """
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.BK,
                           block=qrl_pb2.Block(),
                           mrData=qrllegacy_pb2.MRData()
                           )
        self.manager.handle_block(self.channel, msg)
        self.channel.factory.master_mr.register.assert_called()

        # But if we didn't request this block, we shouldn't process it.
        self.channel.factory.master_mr.register.reset_mock()
        self.channel.factory.master_mr.isRequested.return_value = False

        self.manager.handle_block(self.channel, msg)
        self.channel.factory.master_mr.register.assert_not_called()

    @patch('qrl.core.p2p.p2pChainManager.Block')
    def test_handle_block_bad_block(self, m_Block, m_logger):
        """
        If the block couldn't be constructed from the message, the function should return without doing
        anything else.
        :return:
        """
        m_Block.side_effect = Exception
        msg = make_message(func_name=qrllegacy_pb2.LegacyMessage.BK,
                           block=qrl_pb2.Block(),
                           mrData=qrllegacy_pb2.MRData()
                           )
        self.manager.handle_block(self.channel, msg)
        self.channel.factory.master_mr.register.assert_not_called()
        self.channel.factory.pow.pre_block_logic.assert_not_called()

    def test_handle_block_height_incoming_request(self, m_logger):
        """
        handle_block_height(), unlike other handlers in this class, deals with requests and sends out responses
        to requests, all in the same function!
        If the incoming message.bhData.block_number is 0, this means that the peer wants to know our block height.
        If the incoming message.bhData.block_number is not 0, this means that we should update our knowledge of the
        peer's blockheight.
        """
        self.channel.factory = Mock(
            last_block=Mock(autospec=Block, block_number=5, headerhash=b''),
            get_cumulative_difficulty=Mock(return_value=(0,)))

        incoming_request = make_message(func_name=qrllegacy_pb2.LegacyMessage.BH,
                                        bhData=qrl_pb2.BlockHeightData(block_number=0))

        self.manager.handle_block_height(self.channel, incoming_request)

        self.channel.send.assert_called()
        self.channel.factory.update_peer_blockheight.assert_not_called()

    def test_handle_block_height_incoming_request_but_node_at_blockheight_zero(self, m_logger):
        """
        If the incoming message.bhData.block_number is 0, this means that the peer wants to know our block height.
        However, if we are at blockheight 0 ourselves, we shamefully reply with silence.
        """
        self.channel.factory = Mock(
            last_block=Mock(autospec=Block, block_number=0, headerhash=b''),
            get_cumulative_difficulty=Mock(return_value=(0,)))

        incoming_request = make_message(func_name=qrllegacy_pb2.LegacyMessage.BH,
                                        bhData=qrl_pb2.BlockHeightData(
                                            block_number=0,
                                        ))
        self.manager.handle_block_height(self.channel, incoming_request)

        self.channel.send.assert_not_called()
        self.channel.factory.update_peer_blockheight.assert_not_called()

    def test_handle_block_height_incoming_information(self, m_logger):
        """
        If the incoming message.bhData.block_number is not 0, this means that we should update our knowledge of the
        peer's blockheight.
        """

        some_cumulative_difficulty = bytes(StringToUInt256('0'))
        incoming_info = make_message(func_name=qrllegacy_pb2.LegacyMessage.BH,
                                     bhData=qrl_pb2.BlockHeightData(
                                         block_number=1,
                                         block_headerhash=sha256(b'some_hash'),
                                         cumulative_difficulty=some_cumulative_difficulty
                                     ))
        self.manager.handle_block_height(self.channel, incoming_info)

        self.channel.send.assert_not_called()
        self.channel.factory.update_peer_blockheight.assert_called()

    def test_handle_node_headerhash_peer_has_no_headerhashes(self, m_logger):
        """
        This function also handles requests and responses of NodeHeaderHashes.
        A NodeHeaderHash is a list of blockhashes that a particular node has.
        If the incoming nodeHeaderHash message object has no headerhashes, let's assume the peer needs to sync and send
        a list of our headerhashes.
        """
        self.channel.factory.get_headerhashes.return_value = qrl_pb2.NodeHeaderHash()
        incoming_request = make_message(func_name=qrllegacy_pb2.LegacyMessage.HEADERHASHES,
                                        nodeHeaderHash=qrl_pb2.NodeHeaderHash(block_number=0)
                                        )
        self.manager.handle_node_headerhash(self.channel, incoming_request)
        self.channel.send.assert_called()

    def test_handle_node_headerhash_peer_has_headerhashes(self, m_logger):
        """
        This function also handles requests and responses of NodeHeaderHashes.
        A NodeHeaderHash is a list of blockhashes that a particular node has.
        If the incoming nodeHeaderHash message object has headerhashes, compare our headerhash list and possibly sync.
        """
        incoming_info = make_message(func_name=qrllegacy_pb2.LegacyMessage.HEADERHASHES,
                                     nodeHeaderHash=qrl_pb2.NodeHeaderHash(block_number=3,
                                                                           headerhashes=(b'0', b'1', b'2')))
        self.manager.handle_node_headerhash(self.channel, incoming_info)
        self.channel.factory.compare_and_sync.assert_called()
